package LogA0::CISN;

use strict;
use warnings;
use Math::Trig;

require Exporter;

# set the version for version checking
our $VERSION     = '1.00';

# Inherit from Exporter to export functions and variables
our @ISA         = qw(Exporter);

# Functions and variables which are exported by default
our @EXPORT      = qw(func1 func2);

# Functions and variables which can be optionally exported
our @EXPORT_OK   = qw($Var1 %Hashit func3);

# exported package globals go here

# non-exported package globals go here
# (they are still accessible as $Some::Module::stuff)
#
# Distance limits for CISN LogA0:
our $minDist = 0.1;
our $maxDist = 500.0;
our $badValue = -9.0;
our $linearBreak = 8.0;
# Chebyshev polynomial coefficients:
our @TP = (0.056, -0.031, -0.053, -0.080, -0.028, 0.015);
# Coefficients for Chebyshev mapping:
our $b = 2.0 / (log10($maxDist) - log10($linearBreak));
our $a = -1.0 - $b * log10($linearBreak);

sub new {
    my ($class) = @_;
    my $self = {};

    bless $self, $class;
    return $self;
}

sub minDist {
    my ($self, $dist) = @_;

    return $minDist;
}

sub maxDist {
    my $self = shift;

    return $maxDist;
}

sub logA0 {
    my ($self, $dist) = @_;

    die "logA0 error: distance not defined\n" unless (defined $dist);
    return $badValue if ($dist < $minDist || $dist > $maxDist);

    return linLogA0($dist) if ($dist < $linearBreak);

    # Chebyshev polynomial expansion:
    my $x = z($dist);
    my $mlogA0 = baseLogA0($dist) + 0.0054;
    for (my $j = 0; $j < scalar @TP; $j++) {
	$mlogA0 += $TP[$j] * T( $j + 1, $x );
    }

    return $mlogA0;
}



# Private functions (not exported):
# Compute log base 10:
sub log10 {
    my ($arg) = @_;

    return log($arg) / log(10);
}

# The linear partof the CISN LogA0 function:
sub linLogA0 {
    my ($dist) = @_;

    # Linear extrapolation of avaerage slope between 8 and 60 km:
    my $linB = (2.6182 - 1.5429) / (log10(60) - log10($linearBreak));
    return (1.5429 + $linB * (log10($dist) - log10($linearBreak)));
}

# Map distance (8 - 500) into Chebyshev argument (-1 - 1)
sub z {
    my ($dist) = @_;

    return $a + $b * log10($dist);
}

# The base part of CISN LogA0 curve; the Chebyshev polynomial is added to this:
sub baseLogA0 {
    my ($dist) = @_;

    return 1.11 * log10($dist) + 0.00189 * $dist + 0.591;
}

# Chebyshev function:
sub T {
    my ($n, $x) = @_;

    my $theta = acos( $x );
    return cos( $n * $theta );
}


1;
__END__
# Below is stub documentation for your module. You'd better edit it!

=head1 NAME

LogA0::CISN - Perl extension for the CISN LogA0 function

=head1 SYNOPSIS

  use LogA0::CISN;
  $obj = LogA0::CISN->new();         # Create LogAo::CISN object
  $minDist = $obj->minDist();        # Retrieve the minimum distance for the CISN LogA0 curve
  $maxDist = $obj->maxDist();        # Retrieve the maximum distance for the CISN LogA0 curve
  $myLogA0 = $obj->logA0($dist);     # Return the CISN LogA0 value for specified distance

=head1 DESCRIPTION

Upon successfully reading one event from a hypoinverse archive file, newEvent()
returns a hash reference. (UNDEF is returned otherwise.) The hash may contain
one or more of the following elements:

=over

=item minDist

Returns the minimum distance for which the CISN LogA0 curve is defined.

=item maxDist

Returns the maximum distance for which the CISN LogA0 curve is defined.

=item logA0

Returns the value of the CISN LogA0 curve for the specied distance. If the given distance
is outside the defined limits, returns -9.0.

=back

=head2 EXPORT

None.

=head1 SEE ALSO

BSSA 2011 paper by Bob Uhrhammer at al.

=head1 AUTHOR

Peter Lombard, E<lt>lombard@seismo.berkeley.edu<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2021 by Peter Lombard

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.8 or,
at your option, any later version of Perl 5 you may have available.


=cut
