#!/usr/bin/perl

# Script to collect WA amps for listed events, prepare pairs of Ml values for dMl calcs.

use strict;
use warnings;
use Getopt::Std;
use DBI;
use File::Basename;
use LogA0::CISN;

# Global variables
our($VERSION, $config );
our($opt_c,$opt_e,$opt_h,$opt_s,$opt_S,$opt_B, $opt_O);

# Default configuration file: adjust as necessary
$config = "dMlAppend.conf";

$VERSION = "0.0.2";

##############################################################################
#       print_syntax    - print syntax and exit.
##############################################################################
sub print_syntax {
    my ($cmdname) = @_;
    print STDERR << "EOF";
    $cmdname version $VERSION
        $cmdname - Collect and merge WA amps into pairs for dMl calculations
Syntax:
        $cmdname  [-c config] [-h] -B basis-file (-S SN-file | -s SN) 
                -e event-file [-O output_file]
where:
        -c config     - specify an alternate configuration file
                        The default config file is $config
	-h	      - prints this help message.
	-B basis-file - (required) file listing basis dMl values
        -e event-list - (required) file listing events
        -h            - prints this help message.
	-S SN-file    - file listing station.nets for which dMl values are desired
	-s SN         - station.net for which dMl values are desired.

NOTES:
    ...
EOF
   exit(0);
}


exit main();

# The main function
sub main {

  my ($me) = basename($0);

  # Process command-line arguments
  getopts ('c:e:hB:S:s:O:');

  print_syntax($me) if ($opt_h);
  unless ($opt_e) {
    print STDERR "error: no event list given\n";
    print_syntax($me);
  }
  my $evtFile = $opt_e;
  unless ($opt_B) {
    print STDERR "error: dMl basis list not given\n";
    print_syntax($me);
  }
  my %basisDml;			# hash if dMl basis values by net.sta, dir
  parseBasis($opt_B, \%basisDml);
  unless ($opt_S || $opt_s) {
    print STDERR "error: no SN list or name given\n";
    print_syntax($me);
  }
  if ($opt_S && $opt_s) {
    print STDERR "error: can't use both -S and -s\n";
    print_syntax($me);
  }

  $config = $opt_c if ($opt_c);
  my %params;
  parse_config($config, \%params);
  $params{LOGA0} = LogA0::CISN->new(); # SHould this be configurable?

  my %cand;
  if ($opt_s) {
    $cand{$opt_s} = 1;
  } else {
    parseSN($opt_S, \%cand);
  }

  my $fh;
  if ($opt_O) {
      open($fh, '>', $opt_O) or die "Can't open $opt_O: $!\n";
  } else {
      $fh = \*STDOUT;
  }

  open EVT, "<$evtFile" or die "Can't open $evtFile: $!\n";

  eval {
    $params{DBH} = DBI->connect("dbi:Oracle:" . $params{DBNAME}, $params{DBUSER},
				$params{DBPASS},
				{RaiseError => 1,
				 PrintError => 0,
				 AutoCommit => 0});
    prepareSql(\%params);

    # Process each event in evtFile:
  EVENT:
    while (<EVT>) {
      chomp;
      my @ary = split;
      my $evid = $ary[0];
      my $evtTime = $ary[1];
      my $evtMag = $ary[6];
      my $evtMagType = $ary[7];

      $params{STH}->bind_param(":evid", $evid);
      $params{STH}->execute();
      my $aryref = $params{STH}->fetchall_arrayref({});
      if (0 == scalar @$aryref) {
	print STDERR "no amps found for $evid\n";
	next EVENT;
      }
      my (%basisData, %candData);
    AMP:
      foreach my $ref (@$aryref) {
	my $sta = $ref->{STA};
	my $net = $ref->{NET};
	my $sn = "$sta.$net";
	if (exists $basisDml{$sn}) {
	  storeAmp($ref, \%basisData, $evid);
	} elsif (exists $cand{$sn}) {
	  storeAmp($ref, \%candData, $evid);
	}
	# skip unwanted amp
      }

      # Check amps against various limits, find best amp for SN-DIR
      evalAmps(\%candData, \%params, $evid);
      evalAmps(\%basisData, \%params, $evid);

      # Compute magnitude for good amps
      getMag(\%candData, \%params, $evid, $evtMag);
      getMag(\%basisData, \%params, $evid, $evtMag);

      # Form candidate - basis pairs and write out dMl
      writeDml(\%candData, \%basisData, \%basisDml, \%params, $evid, $fh);
    }
  };
  if ($@) {
    print STDERR "Error: $@\n";
    $params{DBH}->disconnect() if ($params{DBH});
    return 1;
  }

  $params{DBH}->disconnect();
  return 0;
}

sub parse_config {
  my ($config, $pRef) = @_;

  open CF, "< $config" or die "Error opening $config: $!\n";
  while (<CF>) {
    if (/^\s*DB_USER\s*=\s*(\S+)/) {
      $pRef->{DBUSER} = $1;
    } elsif (/^\s*DB_PASSWORD\s*=\s*(\S*)/) {
      $pRef->{DBPASS} = $1;
    } elsif (/^\s*DB_NAME\s*=\s*(\S*)/) {
      $pRef->{DBNAME} = $1;
    } elsif (/^\s*PERIOD_RANGE\s*=\s*(\S*)\s+(\S+)/) {
      $pRef->{PER_MIN} = $1;
      $pRef->{PER_MAX} = $2;
    } elsif (/^\s*BB_AMP_RANGE\s*=\s*(\S*)\s+(\S+)/) {
      $pRef->{BB_MIN} = $1;
      $pRef->{BB_MAX} = $2;
    } elsif (/^\s*SM_AMP_RANGE\s*=\s*(\S*)\s+(\S+)/) {
      $pRef->{SM_MIN} = $1;
      $pRef->{SM_MAX} = $2;
    } elsif (/^\s*AMPRATIO_RANGE\s*=\s*(\S*)\s+(\S+)/) {
      $pRef->{AR_MIN} = $1;
      $pRef->{AR_MAX} = $2;
    } elsif (/^\s*RESIDUAL_MAX\s*=\s*(\S*)/) {
      $pRef->{RES_MAX} = $1;
    } elsif (/^\s*AZIMUTH_LIMIT\s*=\s*(\S*)/) {
      $pRef->{AZI_LIMIT} = $1;
    } elsif (/^\s*DML_LIMIT\s*=\s*(\S*)/) {
      $pRef->{DML_LIMIT} = $1;
    }
  }
  die "DB_USER missing\n" unless (exists $pRef->{DBUSER});
  die "DB_PASSWORD missing\n" unless (exists $pRef->{DBPASS});
  die "DB_NAME missing\n" unless (exists $pRef->{DBNAME});
  die "PERIOD_RANGE missing\n" unless (exists $pRef->{PER_MIN});
  die "BB_AMP_RANGE missing\n" unless (exists $pRef->{BB_MIN});
  die "SM_AMP_RANGE missing\n" unless (exists $pRef->{SM_MIN});
  die "AMPRATIO_RANGE missing\n" unless (exists $pRef->{AR_MIN});
  die "RESIDUAL_LIMIT missing\n" unless (exists $pRef->{RES_MAX});
  die "AZIMUTH_LIMIT missing\n" unless (exists $pRef->{AZI_LIMIT});
  die "DML_LIMIT missing\n" unless (exists $pRef->{DML_LIMIT});

  return;
}

# parse the candidate SN list
# Store in hash by sta.net
sub parseSN {
  my ($file, $ref) = @_;

  open IN, "<$file" or die "Can't open candidate SN list $file: $!\n";
  while (<IN>) {
    chomp;
    next if (/^\s*#/);		# skip comments
    my @ary = split /\./, $_;
    if (2 > scalar @ary) {
      die "Can't find station.net in $_\n";
    }
    my $sn = "$ary[0].$ary[1]";
    $ref->{$sn} = 1;
  }
  close IN;

  return;
}

# Parse file of dMl basis values. dMl values are listed by SNCL, but
# we are interested in SN-direction only. Check that there are not
# multiple values for each SN-direction.
sub parseBasis {
  my ($file, $ref) = @_;

  open IN, "<$file" or die "Can't open dMl basis $file: $!\n";
  while (<IN>) {
    chomp;
    my ($sncl, $dMl) = split;
    my ($s, $n, $c, $l) = split /\./, $sncl;
    my $sn = "$s.$n";
    my $dir = substr $c, 2, 1;
    if (exists $ref->{$sn}->{$dir}) {
      if ($dMl != $ref->{$sn}->{$dir}) {
	die "error: multuple basis dMl values for $sn $dir\n";
      }
    }
    $ref->{$sn}->{$dir} = $dMl;
  }
  close IN;

  return;
}

sub prepareSql {
  my ($pRef) = @_;

  $pRef->{STH} = $pRef->{DBH}->prepare(q{
       SELECT a.sta, a.net, a.seedchan, a.location, a.amplitude, a.per,
       a.datetime, a.amptype, a.units, wheres.separation_km(o.lat, o.lon,
       -o.depth, c.lat, c.lon, c.elev / 1000) DIST, c.azimuth,
       c.samprate, r.gain_units
       FROM event e, eventprefmag ep, origin o, assocamm s, amp a,
       channel_data c, simple_response r
       WHERE e.evid = :evid AND e.prefor = o.orid AND e.evid = ep.evid
       AND ep.magtype = 'l' AND ep.magid = s.magid AND s.ampid = a.ampid
       AND a.net = c.net AND a.sta = c.sta AND a.seedchan = c.seedchan
       AND a.location = c.location
       AND to_date(truetime.getstring(a.datetime), 'YYYY/MM/DD HH24:MI:SS')
       BETWEEN c.ondate and c.offdate AND a.net = r.net AND a.sta = r.sta
       AND a.seedchan = r.seedchan AND a.location = r.location
       AND to_date(truetime.getstring(a.datetime), 'YYYY/MM/DD HH24:MI:SS')
       BETWEEN r.ondate and r.offdate});

  return;
}

# Store one row of amp data retrieved from DB
sub storeAmp {
  my ($row, $data, $evid) = @_;

  my $s = $row->{STA};
  my $n = $row->{NET};
  my $c = $row->{SEEDCHAN};
  my $l = $row->{LOCATION};
  my $sn = "$s.$n";
  my $dir = substr $c, 2, 1;
  $l =~ s/ /-/g;		# replace blanks with dashes
  my $cl = "$c.$l";
  my $sncl = "$sn.$cl";
  my $units = $row->{UNITS};
  my $amp = $row->{AMPLITUDE};
  # Convert amps to millimeters
  if ($units eq "cm") {
    $amp *= 10;
  } elsif ($units eq "mm") {
    ;
  } else {
    print STDERR "Unknown units: $units for $sncl $evid\n";
    return;
  }

  $data->{$sn}->{$dir}->{$cl} = {AMPLITUDE  => $amp,
				 PERIOD     => $row->{PER},
				 AMPTYPE    => $row->{AMPTYPE},
				 DISTANCE   => $row->{DIST},
				 AZIMUTH    => $row->{AZIMUTH},
				 SAMPRATE   => $row->{SAMPRATE},
				 GAIN_UNITS => $row->{GAIN_UNITS},
				 OK   => 1,
				};

  return;
}

# Evaluate amplitudes: are they within limits of amplitude, period, distance?
# If OK, select best amps for each SN-DIR, compute magnitude
sub evalAmps {
  my ($data, $params, $evid) = @_;

 SN:
  foreach my $sn (keys %$data) {
  DIR:
    foreach my $dir (keys %{ $data->{$sn} } ) {
      my $goodAmps = scalar keys %{ $data->{$sn}->{$dir} };
      $data->{$sn}->{$dir}->{SM_OK} = {};
      $data->{$sn}->{$dir}->{BB_OK} = {};
      my $bestCL;
    CL:
      foreach my $cl (keys %{  $data->{$sn}->{$dir} } ) {
	next unless ($cl =~ /^\w{3}\...$/);
	my $ref = $data->{$sn}->{$dir}->{$cl};
	$ref->{IS_SM} = 0;
	# Test Azimuth
	if ($params->{AZI_LIMIT} >= 0) {
	  $ref->{AZIMUTH} -= 180 if ($ref->{AZIMUTH} > 180);
	  if ($dir eq 'N') {
	    if (abs($ref->{AZIMUTH}) > $params->{AZI_LIMIT} &&
		abs($ref->{AZIMUTH} - 180) > $params->{AZI_LIMIT}) {
	      print STDERR "$evid $sn $dir $cl not aligned N/S: $ref->{AZIMUTH}\n";
	      $ref->{OK} = 0;
	      $goodAmps--;
	      next CL;
	    }
	  } elsif ($dir eq 'E') {
	    if (abs($ref->{AZIMUTH} - 90) > $params->{AZI_LIMIT} &&
		abs($ref->{AZIMUTH} - 270) > $params->{AZI_LIMIT}) {
	      print STDERR "$evid $sn $dir $cl not aligned E/W: $ref->{AZIMUTH}\n";
	      $ref->{OK} = 0;
	      $goodAmps--;
	      next CL;
	    }
	  }
	}

	# Test distance
	if ($ref->{DISTANCE} < $params->{LOGA0}->minDist() ||
	    $ref->{DISTANCE} > $params->{LOGA0}->maxDist()) {
	  print STDERR "$evid $sn $dir $cl out of range for CISN LogA0: $ref->{DISTANCE}\n";
	  $ref->{OK} = 0;
	  $goodAmps--;
	  next CL;
	}

	# Test period
	if ($ref->{PERIOD} < $params->{PER_MIN} ||
	    $ref->{PERIOD} > $params->{PER_MAX}) {
	  print STDERR "$evid $sn $dir $cl period outside bounds: $ref->{PERIOD}\n";
	  $ref->{OK} = 0;
	  $goodAmps--;
	  next CL;
	}

	# Last CL test; save good CLs
	# Test amplitude range
	if ($ref->{GAIN_UNITS} eq "DU/M/S**2") {
	  if ($ref->{AMPLITUDE} < $params->{SM_MIN} ||
	      $ref->{AMPLITUDE} > $params->{SM_MAX} ) {
	    print STDERR "$evid $sn $dir $cl amplitude outside SM bounds: $ref->{AMPLITUDE}\n";
	    $ref->{OK} = 0;
	    $goodAmps--;
	    next CL;
	  }
	  $ref->{IS_SM} = 1;
	  $data->{$sn}->{$dir}->{SM_OK}->{$cl} = 1;
	} else {
	  if ($ref->{AMPLITUDE} < $params->{BB_MIN} ||
	      $ref->{AMPLITUDE} > $params->{BB_MAX} ) {
	    print STDERR "$evid $sn $dir $cl amplitude outside BB bounds: $ref->{AMPLITUDE}\n";
	    $ref->{OK} = 0;
	    $goodAmps--;
	    next CL;
	  }
	  $data->{$sn}->{$dir}->{BB_OK}->{$cl} = 1;
	}
      }

      # If we have both BB and SM amps at this point, test their ratios
      if (scalar keys %{ $data->{$sn}->{$dir}->{BB_OK} } &&
	  scalar keys %{ $data->{$sn}->{$dir}->{SM_OK} } ) {
	testRatios($data->{$sn}->{$dir}, $params, $sn, $evid);
      }
      my @OK = keys %{ $data->{$sn}->{$dir}->{BB_OK} };
      push @OK, keys %{ $data->{$sn}->{$dir}->{SM_OK} };
      if (0 == scalar @OK) {
	print STDERR "$evid no good amps for $sn $dir\n";
	next DIR;
      } elsif (1 == scalar @OK) {
	print STDERR "$evid 1 good amp for $sn $dir\n";
	$bestCL =  $OK[0];
      } else {
	printf STDERR "%8d %d good amps for %s %s %s;",
	  $evid, scalar @OK, $sn, $dir, join ", ", sort @OK;

	# Find best CL:
	my $orderCL = getCmp( $data->{$sn}->{$dir});
	my @sortCL = sort $orderCL @OK;
	$bestCL = $sortCL[0];
	print STDERR " best is $bestCL\n";
      }

      $data->{$sn}->{$dir}->{BEST} = $bestCL;
    }
  }

  return;
}


# Generate a comparison function which can access the given reference
# to $data->{$sn}->{$dir}, i.e. all the CL info.
# Strong-motion preferred over broadband, higher sample rate over lower,
# and alphabetically lower location code.
sub getCmp {
  my $ref = shift;

  return sub {
    my $la = substr $a, 4, 2;
    my $lb = substr $b, 4, 2;
    return (($ref->{$b}->{IS_SM} <=> $ref->{$a}->{IS_SM}) ||
      ($ref->{$b}->{SAMPRATE} <=> $ref->{$a}->{SAMPRATE}) ||
	$la cmp $lb);
  }
}

# Compare BB and SM amplitude ratios for a given SN-DIR.
# Based on Bob's amp_ratio subroutine in merge_was_stn_eq_data.f.
# His code discards the SM amp if the amp ratio is outside limits; why?
sub testRatios {
  my ($ref, $params, $sn, $evid) = @_;

  my @badSM;
  foreach my $bb (keys %{ $ref->{BB_OK} } ) {
    foreach my $sm (keys %{ $ref->{SM_OK} } ) {
      my $ar = $ref->{$bb}->{AMPLITUDE} / $ref->{$sm}->{AMPLITUDE};
      if ($ar < $params->{AR_MIN} || $ar > $params->{AR_MAX}) {
	print STDERR "$evid $sn amp ratio for $bb / $sm: $ar outside limits\n";
	push @badSM, $sm;
      }
    }
  }

  foreach my $sm (@badSM) {
    delete $ref->{SM_OK}->{$sm};
    $ref->{$sm}->{OK} = 0;
  }

  return;
}


# Compute Ml magnitude for the best CL of each SN-Dir.
# Check that the residual falls within limits
sub getMag {
  my ($data, $params, $evid, $evtMag) = @_;

  foreach my $sn (keys %$data ) {
    foreach my $dir (keys %{ $data->{$sn} } ) {
      if (exists $data->{$sn}->{$dir}->{BEST}) {
	my $ref = $data->{$sn}->{$dir}->{$data->{$sn}->{$dir}->{BEST}};
	my $ml = log10($ref->{AMPLITUDE}) + $params->{LOGA0}->logA0($ref->{DISTANCE});
	# DEBUG
	# printf STDERR "%8d %15s %10.6f %8.6f %8.6f\n", $evid, "$sn.$dir",
	# $ref->{DISTANCE}, log10($ref->{AMPLITUDE}), 
	# $params->{LOGA0}->logA0($ref->{DISTANCE});

	if (abs($ml - $evtMag) > $params->{RES_MAX}) {
	  print STDERR "$evid $sn $data->{$sn}->{$dir}->{BEST} exceeds residual limit: $ml\n";
	  $ref->{OK} = 0;
	  delete $data->{$sn}->{$dir}->{BEST};
	  next;
	}
	$ref->{ML} = $ml;
      }
    }
  }

  return;
}

# Form pairs of SNCL-dMl values between candidate and basis stations,
# write out event dMl values
sub writeDml {
  my ($cand, $basis, $dml, $params, $evid, $fh) = @_;

  foreach my $cSn (sort keys %$cand ) {
    foreach my $cDir (sort keys %{ $cand->{$cSn} } ) {
      next unless (exists $cand->{$cSn}->{$cDir}->{BEST});
      my $cRef = $cand->{$cSn}->{$cDir}->{$cand->{$cSn}->{$cDir}->{BEST}};
      my $cSncl ="$cSn.$cand->{$cSn}->{$cDir}->{BEST}";
      foreach my $bSn (sort keys %$basis) {
	foreach my $bDir (sort keys %{ $basis->{$bSn} } ) {
	  next unless (exists $basis->{$bSn}->{$bDir}->{BEST});
	  my $bRef = $basis->{$bSn}->{$bDir}->{$basis->{$bSn}->{$bDir}->{BEST}};
	  my $bSncl = "$bSn.$basis->{$bSn}->{$bDir}->{BEST}";
	  my $newDml = $bRef->{ML} + $dml->{$bSn}->{$bDir} - $cRef->{ML};
	  if (abs($newDml) < $params->{DML_LIMIT}) {
	      printf $fh "%8d %15s %7.5f\n", $evid, $cSncl, $newDml;
	  }
	  else {
	      print STDERR "$evid $cSncl $bSncl exceeds dMl limit: $newDml\n";
	  }
	}
      }
    }
  }

  return;
}

sub log10 {
  my ($arg) = @_;

  return log($arg) / log(10);
}
