#!/usr/bin/perl

use strict;
use warnings;
use DBI;
use File::Basename;
use Getopt::Std;
use LogA0::CISN;

our ($config, $dbh, $dbname, $dbuser, $dbpass, $VERSION);
our ($opt_c, $opt_h);

$config = "dMlAppend.conf";
$VERSION = "0.0.1";

##############################################################################
#       print_syntax    - print syntax and exit.
##############################################################################
sub print_syntax {
    my ($cmdname) = @_;
    print STDERR << "EOF";
    $cmdname version $VERSION
        $cmdname - Check amps for one sta.net
Syntax:
        $cmdname  [-c config] [-h] sta.net event-file [amp_merge_screen]
where:
        -c config     - specify an alternate configuration file
                        The default config file is $config
	-h            - prints this help message.
	sta.net	      - station and net code to be checked
	event-file    - file listing events examined for this sta.net
	amp_merge_screen - (optional) "screen" output of amp_merge program

NOTES:
Given a sta.net and an event list, query DB for SN's amps for those events,
show sta mags and residuals to identify problem amps.
If amp_merge.pl output is provided, shows any amps rejected by amp_merge.pl

    ...
EOF
   exit(0);
}

exit main();

# The main function
sub main {

  my ($me) = basename($0);

  # Process command-line arguments
  getopts ('c:h');

  print_syntax($me) if ($opt_h);
  print_syntax($me) unless (1 < scalar @ARGV);

  $config = $opt_c if ($opt_c);
  parse_config($config);

  my $haveScreen = 0;
  my %screen;
  my $sn = $ARGV[0];
  my ($s, $n) = split /\./, $sn;
  my $evtFile = $ARGV[1];
  if (3 == scalar @ARGV) {
      $haveScreen = 1;
      readScreen($ARGV[2], $sn, \%screen);
  }

  my $logA0 = LogA0::CISN->new();

  open EVT, "<$evtFile" or die "Can't open $evtFile: $!\n";
  print "EVID      SNCL           eMag dist  sMag resid\n";

  eval {
      $dbh = DBI->connect("dbi:Oracle:" . $dbname, $dbuser, $dbpass,
			  {RaiseError => 1,
			   PrintError => 0,
			   AutoCommit => 0});
      my $sth = $dbh->prepare(q{
       SELECT n.magnitude, a.amplitude, a.units, a.seedchan, a.location,
       wheres.separation_km(o.lat, o.lon, -o.depth, c.lat, c.lon,
       c.elev / 1000) DIST
       FROM event e, eventprefmag ep, origin o, netmag n, assocamm s, amp a,
       channel_data c
       WHERE e.evid = :evid AND e.evid = ep.evid AND ep.magtype = 'l'
       AND ep.magid = n.magid AND ep.magid = s.magid AND s.ampid = a.ampid
       AND a.net = :net AND a.sta = :sta AND e.prefor = o.orid
       AND a.net = c.net AND a.sta = c.sta AND a.seedchan = c.seedchan
       AND a.location = c.location
       AND to_date(truetime.getstring(a.datetime), 'YYYY/MM/DD HH24:MI:SS')
       BETWEEN c.ondate and c.offdate});

      while (<EVT>) {
	  chomp;
	  my @ary = split;
	  my $evid = $ary[0];
	  $sth->bind_param(":evid", $evid);
	  $sth->bind_param(":net", $n);
	  $sth->bind_param(":sta", $s);
	  $sth->execute();
	  my $aryref = $sth->fetchall_arrayref({});
	  unless ($aryref && 0 < scalar @$aryref) {
	      print STDERR "no amps for $evid\n";
	      next;
	  }
	  for my $ref (@$aryref) {
	      my $chan = $ref->{SEEDCHAN};
	      my $loc = $ref->{LOCATION};
	      $loc =~ s/ /-/g;
	      my $cl = "$chan.$loc";
	      my $amp = $ref->{AMPLITUDE};
	      my $units = $ref->{UNITS};
	      # Convert amps to millimeters
	      if ($units eq "cm") {
		  $amp *= 10;
	      }
	      my $sMag = $ref->{mag};
	      my $eMag = $ref->{MAGNITUDE};
	      my $dist = $ref->{DIST};
	      unless (defined $sMag && $dist >= $logA0->minDist() &&
			  $dist <= $logA0->maxDist()) {
		  $sMag = staMag($amp, $dist, $logA0);
	      }
	      my $sMagStr = sprintf("%4.2f", $sMag);
	      my $magRes = sprintf("%5.2f", $sMag - $eMag);

	      printf "%8d %5s.%2s.%3s.%2s %4.2f %5.1f %4s %5s", $evid, $s, $n,
		  $chan, $loc, $eMag, $dist, $sMagStr, $magRes;
	      if ($haveScreen && exists $screen{$evid} ) {
		  my $reason = findReason($evid, $cl, \%screen);
		  print " dropped: $reason" if (defined $reason);
	      }
	      print "\n";
	  }
	  print "\n";  # to separate events
      }
      close EVT;
  };
  if ($@) {
      print STDERR "error: $@\n";
  }
  $dbh->disconnect() if ($dbh);

  return 0;
}

sub staMag {
  my ($amp, $dist, $logA0) = @_;

  return log10($amp) + $logA0->logA0($dist);
}

sub log10 {
  my ($arg) = @_;

  return log($arg) / log(10);
}

sub readScreen {
  my ($file, $sn, $ref) = @_;

  open IN, "<$file" or die "Can't open screen $file: $!\n";
  while (<IN>) {
    chomp;
    my @ary = split;
    if ($ary[1] eq $sn) {
      my $evid = $ary[0];
      my $cl = $ary[3];
      my $reason;
      if (/aligned/) {
	$reason = "azimuth";
      }
      elsif (/range/) {
	$reason = "distance";
      }
      elsif (/period/) {
	$reason = "period";
      }
      elsif (/amplitude/) {
	$reason = "amplitude";
      }
      elsif (/ratio/) {
	$cl = "$ary[5] - $ary[7]";
	$reason = "amp ration";
      }
      else {
	$reason = "unknown";
      }
      $ref->{$evid}->{$cl} = $reason;
    }
  }
  close IN;

  return;
}

sub findReason {
  my ($evid, $cl, $ref) = @_;

  my $reason;
  if (exists $ref->{$evid}->{$cl}) {
    $reason = $ref->{$evid}->{$cl};
  }
  else {
    foreach my $key (keys %{ $ref->{$evid} } ) {
      if ($key =~ /$cl/) {
	$reason = $ref->{$evid}->{$key};
	last;
      }
    }
  }

  return $reason;
}

sub parse_config {
  my ($config) = @_;

  open CF, "< $config" or die "Error opening $config: $!\n";
  while (<CF>) {
      if (/^\s*DB_USER\s*=\s*(\S+)/) {
	  $dbuser = $1;
      } elsif (/^\s*DB_PASSWORD\s*=\s*(\S*)/) {
	  $dbpass = $1;
      } elsif (/^\s*DB_NAME\s*=\s*(\S*)/) {
	  $dbname = $1;
      }
  }
  close CF;

  die "DB_USER missing\n" unless (defined $dbuser);
  die "DB_PASSWORD missing\n" unless (defined $dbpass);
  die "DB_NAME missing\n" unless (defined $dbname);

  return;
}
