#!/usr/bin/perl

# "simple" script to compute dMl statistics from output of 
# "new_cisn_sncl_dml_calc" program
#
# Attempts to compute dMl values for North and East components of station.net,
# merging dMls from staton's channels and location codes.
# This seems to be how NCSS, at least, handles dMls.
# Internally this code deals in NSCL to make sorting easier, but
# writes output in SNCL to conform to older codes

use strict;
use warnings;

unless (2 == scalar @ARGV) {
    print "usage: $0 sncl_dml_file sncl_data_file\n";
    exit 1;
}

my $inFile = $ARGV[0];
my $staFile = $ARGV[1];

my (%data, %station);

# Read input files:
readStationFile($staFile, \%station);
readDataFile($inFile, \%data);

# Collect dMl values by NS:
collectdMl(\%data, \%station);

# compute dMl statistics
computeStats(\%data);

# Write results to STDOUT
writeOut(\%data);

exit 0;

# Read SNCLs and their sample rate from CISN_Candidate_SNCL_List
sub readStationFile {
    my ($file, $staRef) = @_;
    
    open IN, "<$file" or die "readStationFile: can't open $file: $!\n";
    while (<IN>) {
	chomp;
	my @ary = split;
	die "Can't parse $file\n" unless (8 <= scalar @ary);
	my $nscl = "$ary[1].$ary[0].$ary[2].$ary[3]";
	$staRef->{$nscl} = $ary[7];
    }
    close IN;
    
    return;
}

# Read dml data file from new_cisn_sncl_dml_calc
# Collect data by sta.net, direction, event, chan.loc in array.
# File format: yyyymmddhhmmss SNCL raw-dMl
sub readDataFile {
    my ($file, $data) = @_;

    open IN, "<$file" or die "readDataFile: Can't open $file: $!\n";
    while (<IN>) {
	chomp;
	my ($event, $sncl, $dMl) = split;
	my ($sta, $net, $chan, $loc) = split /\./, $sncl;
	my $ns = "$net.$sta";
	my $cl = "$chan.$loc";
	my $dir = substr $chan, 2, 1;
	unless (exists $data->{$ns}->{$dir}->{EVENT}->{$event}->{$cl}) {
	    $data->{$ns}->{$dir}->{EVENT}->{$event}->{$cl} = [];
	}
	push @{ $data->{$ns}->{$dir}->{EVENT}->{$event}->{$cl} }, $dMl;
    }
    close IN;

    return;
}

# For each NS and direction in $data hash-ref, collect appropriate dMl
# values into an array from the event hash-refs.
# Also determine the dominant CL for this NS and direction.
sub collectdMl {
    my ($data, $staInfo) = @_;

    foreach my $ns (keys %$data) {
	foreach my $dir (keys %{ $data->{$ns} } ) {
	    my $dirD = $data->{$ns}->{$dir};
	    $dirD->{DMLARY} = [];
	    $dirD->{NEQ} = 0;
	    $dirD->{CL} = {};

	    foreach my $evt ( keys %{ $dirD->{EVENT} } ) {
		my $evref = $dirD->{EVENT}->{$evt};
		$dirD->{NEQ}++;
		my @cl = keys %$evref;
		if (1 == scalar @cl) {
		    $dirD->{CL}->{$cl[0]} = 1;
		    push @{ $dirD->{DMLARY} }, @{ $evref->{$cl[0]} };
		}
		else {
		    my $bestCl = findBestCl( $staInfo, $ns, @cl);
		    $dirD->{CL}->{$bestCl} = 1;
		    push @{ $dirD->{DMLARY} }, @{ $evref->{$bestCl} };
		    print STDERR "$evt $ns $dir multiple chans: ";
		    foreach my $cl (sort keys %$evref) {
			print STDERR "$cl $evref->{$cl} ";
		    }
		    print STDERR "using $bestCl\n";
		}
	    }
	    # Could delete EVENT here
	}
    }
    
    return;
}

# For each NS and direction, compute median dMl and other stats
sub computeStats {
    my $data = shift;

    foreach my $ns (keys %$data) {
	foreach my $dir (keys %{ $data->{$ns} } ) {
	    my $dirD = $data->{$ns}->{$dir};
	    my @dMl = sort {$a <=> $b } @{ $dirD->{DMLARY} };
	    my $n = scalar @dMl;
	    $dirD->{NOBS} = $n;
	    next if ($n < 4);
	    my $nq1 = int($n / 4);
	    my $nq2 = int($n / 2);
	    my $nq3 = int($n * 0.75);
	    my $q1 = $dMl[$nq1 - 1];
	    my $median = $dMl[$nq2 - 1];
	    my $q3 = $dMl[$nq3 - 1];
	    my $se = ($q3 - $q1) / sqrt($nq3 - $nq1);
	    $dirD->{DML} = $median;
	    $dirD->{SE} = $se;
	}
    }

    return;
}

sub writeOut {
    my $data = shift;

    foreach my $ns (sort keys %$data) {
	my ($net, $sta) = split /\./, $ns;
	foreach my $dir (sort keys %{ $data->{$ns} } ) {
	    my $ref = $data->{$ns}->{$dir};
	    foreach my $cl (sort keys %{ $ref->{CL} } ) {
		my $sncl = "$sta.$net.$cl";
		printf "%15s %6.3f %6.3f %5d %3d\n", $sncl, $ref->{DML},
		$ref->{SE}, $ref->{NOBS}, $ref->{NEQ};
	    }
	}
    }

    return;
}

sub findBestCl {
    my $staInfo = shift;
    my $ns = shift;
    my @cl = @_;

    # Prefer Strong-motion channels:
    my @sm = grep {/^.[GLN]/} @cl;
    if (1 == scalar @sm) {
	return $sm[0];
    }
    elsif (@sm) {
	my @bestSR = highestSR($staInfo, $ns, @sm);
	if (1 == scalar @bestSR) {
	    return $bestSR[0];
	}
	else {
	    my @bestLoc = bestLoc(@sm);
	    return $bestLoc[0];
	}
    }

    # all are BB chans
    if (1 == scalar @cl) {
	return $cl[0];
    }
    elsif (@cl) {
	my @bestSR = highestSR($staInfo, $ns, @cl);
	if (1 == scalar @bestSR) {
	    return $bestSR[0];
	}
	else {
	    my @bestLoc = bestLoc(@cl);
	    return $bestLoc[0];
	}
    }
    
    # should never get here!
    die "No best CL found for $ns " . join( " ", @cl) . "\n";
}

sub highestSR {
    my $staInfo = shift;
    my $ns = shift;
    my @cl = @_;

    my $bestSR = 0;
    my @bestCL;
    foreach my $cl (@cl) {
	my $nscl = "$ns.$cl";
	die "Unknown NSCL $nscl\n" unless (exists $staInfo->{$nscl});
	my $sr = $staInfo->{$nscl};
	if ($sr > $bestSR) {
	    $bestSR = $sr;
	    undef @bestCL;
	    push @bestCL, $cl;
	}
	elsif ($sr == $bestSR) {
	    push @bestCL, $cl;
	}
    }

    return @bestCL;
}

# Choose the "lowest" location code
sub bestLoc {
    my @cl = @_;

    my $bestLoc = "ZZ";
    my @bestCL;
    foreach my $cl (@cl) {
	my ($c, $l) = split /\./, $cl;
	if ($l lt $bestLoc) {
	    undef @bestCL;
	    push @bestCL, $cl;
	}
	elsif ($l eq $bestLoc) {
	    push @bestCL, $cl;
	}
    }

    return @bestCL;
}


# Probably don't need this; write dMls for all CLs instead of this.
# Generate an appropriate NSCL for given NS
# $data is reference to the appropriate DIR hash of NS.
# This $data hash-reference is expected to have a DOM_CL element already created.
sub genNscl {
    my ($ns, $data) = @_;

    die "genNscl: missing DOM_CL for $ns\n"
	unless (exists $data->{DOM_CL});
    my $nscl = "$ns." . $data->{DOM_CL};

    return $nscl;
}
