#!/usr/bin/perl

# Script to print stations' dMl values when at least
# one direction's dMl has sufficient events contributing.
# Reads output of dml_stats.pl

use strict;
use warnings;
use Getopt::Std;
use File::Basename;

# Global variables
our($VERSION );
our($opt_E, $opt_e, $opt_h, $opt_s);

$VERSION = "1.1 2021/10/02";

##############################################################################
#       print_syntax    - print syntax and exit.
##############################################################################
sub print_syntax {
    my ($cmdname) = @_;
    print STDERR << "EOF";
    $cmdname version $VERSION
        $cmdname - Print stations\' dMl values when they meet specified
		criteria

Syntax:
        $cmdname  [-h] -E nEQ1 [-e nEQ2 -s SE] new_CISN_SNCL_dML.solution
where:
        -E nEQ1       - (required) dMl must have at least nEQ1 events
        -e nEQ2       - one station\'s dMl must have nEQ2 events plus specified
			standard error
	-s SE         - (required when -e given) maximum Standard Error (SE)
        -h            - prints this help message.
	new_CISN_SNCL_dML.solution file to be scanned for acceptable dMls

NOTES:
    If dMl values for both directions meet the nEQ1 threshold, the station is
	accepted.
    If the dMl value for only one direction meets the nEQ1 threshold, the other
	direction\'s dMl must meet the nEQ2 and SE threshold for the station
	to be accepted.
    If only nEQ1 is provided, dMl values for both direction must meet the nEQ1
	threshold for the station to be accepted.

EOF
    exit(0);
}

    exit main();

# The main function
sub main {

    my ($me) = basename($0);

    # Process command-line arguments
    getopts ('E:e:hs:');

    print_syntax($me) if ($opt_h || 1 > scalar @ARGV);
    print_syntax($me) unless ($opt_E);
    my $allowMarginal = 0;
    if ($opt_e && $opt_s) {
	$allowMarginal = 1;
    } elsif ($opt_e || $opt_s) {
	print_syntax($me);
    }

    my $inFile = $ARGV[0];

    my %data;
    my $lastSN = "";
    my %isGood = (N => 0,
		  E => 0,
		 );
    # Grade assigned to $isGood:
    # If a dMl value has nEQ >= nEQ1 ($opt_E), add 2 to $isGood
    # But if the dMl value only has nEQ >= nEQ2 ($opt_e) it must also 
    # have SE <= $opt_s and then gets 1 added to $isGood
    # Since a station must have at least one dMl that meets nEQ1 rule but
    # may have the second dMl meeting nEQ2 and SE rules, the station
    # is accepted if $isGood is >= 3.
    open IN, "<$inFile" or die "Can't open $inFile: $!\n";
    while (<IN>) {
	chomp;
	my ($sncl, $dMl, $se, $nobs, $neq) = split;
	my ($sta, $net, $chan, $loc) = split /\./, $sncl;
	my $sn = "$sta.$net";
	my $dir = substr $chan, 2, 1;
	unless (exists $isGood{$dir}) {
	    $isGood{$dir} = 0;
	}
	if ($lastSN && $sn ne $lastSN) {
	    if (isGood(\%isGood)) {
		foreach my $dir (sort keys %data) {
		    foreach my $ref (@{ $data{$dir} } ) {
			printf "%15s %6.3f %6.3f %5d %3d%s\n", $ref->{SNCL}, 
			    $ref->{DML}, $ref->{SE}, $ref->{NOBS}, $ref->{NEQ},
				(($isGood{$dir} > 1) ? "" : "  *");
		    }
		}
	    }
	    undef %data;
	    $lastSN = $sn;
	    %isGood = (N => 0,
		       E => 0,
		      );
	}
	$data{$dir} = [] unless exists $data{$dir};
	push @{ $data{$dir} }, {
	    SNCL => $sncl,
	    DML  => $dMl,
	    SE   => $se,
	    NOBS => $nobs,
	    NEQ  => $neq,
	};
	if ($neq >= $opt_E) {
	    $isGood{$dir} = 2;
	} elsif ($allowMarginal) {
	    if ($neq >= $opt_e && $se <= $opt_s) {
		$isGood{$dir} = 1 unless ($isGood{$dir} == 2);
	    }
	}
    }
    if (isGood(\%isGood)) {
	foreach my $dir (sort keys %data) {
	    foreach my $ref (@{ $data{$dir} } ) {
		printf "%15s %6.3f %6.3f %5d %3d%s\n", $ref->{SNCL}, $ref->{DML},
		    $ref->{SE}, $ref->{NOBS}, $ref->{NEQ},
			(($isGood{$dir} > 1) ? "" : "  *");
	    }
	}
    }

    return 0;
}

sub isGood {
    my $good = shift;
    my $sncl = shift;

    my $nDir = scalar keys %$good;
    if (2 < $nDir) {
	print STDERR "$sncl has excessive directions: $nDir\n";
	return 0;
    }
    my $count = 0;
    foreach my $key (keys %$good) {
	$count += ($good->{$key});
    }

    return ($count > 2);
}
