#!/usr/bin/perl

use strict;
use warnings;
use File::Basename;
use Getopt::Std;
use DBI;
use Text::CSV_XS;

# Global variables
our ( $VERSION, $config, $dbname, $dbuser, $dbpass, $dbh );
our ($opt_c, $opt_h);

# Default configuration file: adjust as necessary
$config = "dMlAppend.conf";
$VERSION = "1.00 (2022/12/01)";

##############################################################################
#       print_syntax    - print syntax and exit.
##############################################################################
sub print_syntax {
    my ($cmdname) = @_;
    print STDERR << "EOF";
    $cmdname version $VERSION
        $cmdname - Convert dMl stats text file to CSV format
Syntax:
        $cmdname  [-c config] [-h] dml_stats
where:
        -c config     - specify an alternate configuration file
                        The default config file is $config
	-h            - prints this help message.
	dml_stats     - text file with dMl stats

NOTES:
Writes to STDOUT in CSV format.
    ...
EOF
    exit(0);
}

    exit main();

# The main function
sub main {
    my ($me) = basename($0);

    # Process command-line arguments
    getopts ('c:h');

    print_syntax($me) if ($opt_h);
    print_syntax($me) unless (1 == scalar @ARGV);

    $config = $opt_c if ($opt_c);
    parse_config($config);

    my $file = $ARGV[0];

    eval {
	open IN, "<$file" or die "Can't open $file: $!\n";

	$dbh = DBI->connect("dbi:Oracle:" . $dbname, $dbuser, $dbpass,
			    {RaiseError => 1,
			     PrintError => 0,
			     AutoCommit => 0});

	my $sth = $dbh->prepare(q{
           SELECT UNIQUE lat, lon, edepth, azimuth, dip,
           TO_CHAR(ondate, 'YYYY/MM/DD') ONDATE,
           TO_CHAR(offdate, 'YYYY/MM/DD') OFFDATE
           FROM channel_data WHERE sta = :sta AND net = :net
           AND seedchan = :chan AND location = :loc
           ORDER BY ONDATE DESC});

	my %data;
      SNCL:
	while (<IN>) {
	    chomp;
	    next if (/^\s*#/);	# skip comments
	    my ($sncl, $dMl, $se, $nobs, $neq)  = split;
	    my ($s, $n, $c, $l) = split /\./, $sncl;
	    my $ns = "$n.$s";	# facilitates sorting by net then station
	    my $ldb = ($l eq "--") ? "  " : $l; # location code for DB
	    my $dir = substr $c, 2, 1;		# direction code
	    unless ($dir =~ /^[EN]$/) {
		print STDERR "Unexpected direction code for $sncl; skipping\n";
		next SNCL;
	    }
	    $data{$ns} = {} unless (exists $data{$ns});
	    if (exists $data{$ns}->{$dir}) {
		if ( $data{$ns}->{$dir}->{DML} != $dMl) {
		    print STDERR "Warning: multiple dMl values for $ns $dir\n";
		}
		next SNCL;
	    } else {
		# Get metadata
		$sth->bind_param(":net", $n);
		$sth->bind_param(":sta", $s);
		$sth->bind_param(":chan", $c);
		$sth->bind_param(":loc", $ldb);
		$sth->execute();
		my $aryref = $sth->fetchall_arrayref({});
		my $nrow = scalar @$aryref;
		die "No channel_data found for $sncl\n" if ($nrow == 0);
		my $lat = $aryref->[0]->{LAT};
		my $lon = $aryref->[0]->{LON};
		my $depth = $aryref->[0]->{EDEPTH};
		my $azi = $aryref->[0]->{AZIMUTH};
		my $dip = $aryref->[0]->{DIP};
		my $offdate = $aryref->[0]->{OFFDATE};
		my $ondate = $aryref->[-1]->{ONDATE};
		my ($latCh, $lonCh, $dipCh, $aziCh) = (0,0,0,0);
		for (my $i = 1; $i < $nrow; $i++) {
		    if (abs($aryref->[$i]->{LAT} - $lat) > 0.0001) {
			print STDERR "$sncl latitude changed; using latest\n"
			    unless ($latCh);
			$latCh = 1;
		    }
		    if (abs($aryref->[$i]->{LON} - $lon) > 0.0001) {
			print STDERR "$sncl longitude changed; using latest\n"
			    unless ($lonCh);
			$lonCh = 1;
		    }
		    if ($aryref->[$i]->{DIP} != $dip) {
			print STDERR "$sncl dip changed; using latest\n"
			    unless ($dipCh);
			$dipCh = 1;
		    }
		    if (abs($aryref->[$i]->{AZIMUTH} - $azi) > 1 &&
			    abs($aryref->[$i]->{AZIMUTH} - $azi) < 179) {
			print STDERR "$sncl azimuth changed; using latest\n"
			    unless ($aziCh);
			$aziCh = 1;
		    }
		}
		$data{$ns}->{$dir} = {DML   => $dMl,
				      SE    => $se,
				      NOBS  => $nobs,
				      NEQ   => $neq,
				      LAT   => $lat,
				      LON   => $lon,
				      DEPTH => $depth,
				      DIP   => $dip,
				      AZI   => $azi,
				      ON    => $ondate,
				      OFF   => $offdate,
				  };
	    }
	}
	close IN;
	$dbh->disconnect();

	my $csv = Text::CSV_XS->new( {eol => "\n",} );
	my @head = (qw{STATION NET LAT LON DEPTH});
	push @head, "ON DATE", "OFF DATE", "N AZI", "N DIP", "N dMl", "N SE",
	    "N nOBS", "N nEQ", "E AZI", "E DIP", "E dMl", "E SE", "E nOBS",
		"E nEQ";
	$csv->print( *STDOUT, \@head);
	foreach my $ns (sort keys %data) {
	    my ($n, $s) = split /\./, $ns;
	    # We assume N and E components have same lat, lon, depth, dates!
	    my @row = ($s, $n, $data{$ns}->{N}->{LAT}, $data{$ns}->{N}->{LON},
		       $data{$ns}->{N}->{DEPTH}, $data{$ns}->{N}->{ON},
		       $data{$ns}->{N}->{OFF}, $data{$ns}->{N}->{AZI},
		       $data{$ns}->{N}->{DIP}, $data{$ns}->{N}->{DML},
		       $data{$ns}->{N}->{SE}, $data{$ns}->{N}->{NOBS},
		       $data{$ns}->{N}->{NEQ}, $data{$ns}->{E}->{AZI},
		       $data{$ns}->{E}->{DIP}, $data{$ns}->{E}->{DML},
		       $data{$ns}->{E}->{SE}, $data{$ns}->{E}->{NOBS},
		       $data{$ns}->{E}->{NEQ});
	    $csv->print( *STDOUT, \@row);
	}

    };
    if ($@) {
	print STDERR "Error: $@\n";
	$dbh->disconnect() if ($dbh);
    }

    return 0;
}

sub parse_config {
    my ($config) = @_;

    open CF, "< $config" or die "Error opening $config: $!\n";
    while (<CF>) {
	if (/^\s*DB_USER\s*=\s*(\S+)/) {
	    $dbuser = $1;
	} elsif (/^\s*DB_PASSWORD\s*=\s*(\S*)/) {
	    $dbpass = $1;
	} elsif (/^\s*DB_NAME\s*=\s*(\S*)/) {
	    $dbname = $1;
	}
    }
    close CF;

    die "DB_USER missing\n" unless (defined $dbuser);
    die "DB_PASSWORD missing\n" unless (defined $dbpass);
    die "DB_NAME missing\n" unless (defined $dbname);

    return;
}
