#!/usr/bin/perl

# Script to run dMl codes on list of stations

use strict;
use warnings;
use File::Basename;
use File::Copy;
use Getopt::Std;

our ($config, $VERSION, $endDate, $gridSpace, $beforeTime, $afterTime, $minEvents);
our ($nEQ1, $nEQ2, $maxSE, $bin);
our ($opt_c, $opt_h);

# Add locations of LogA0-CISN and Qtime modules if not in standard locations:
unless ($ENV{PERL5LIB} =~ /tulainyo/) {
    $ENV{PERL5LIB} = "/data/tulainyo/lombard/lib/perl5:$ENV{PERL5LIB}";
}

# Default configuration file: adjust as necessary
$config = "dMlAppend.conf";
$VERSION = "1.0.3 (2022/12/02)";

##############################################################################
#       print_syntax    - print syntax and exit.
##############################################################################
sub print_syntax {
    my ($cmdname) = @_;
    print STDERR << "EOF";
    $cmdname version $VERSION
        $cmdname - Collect and merge WA amps into pairs for dMl calculations
Syntax:
        $cmdname  [-c config] [-h] sta-net-date-list
where:
        -c config     - specify an alternate configuration file
                        The default config file is $config
	-h	      - prints this help message.
	sta-net-date-list  - file listing sta.net and start date for which dMl values are needed.
			Start date is in "YYYY/MM/DD" format (without quotes).

NOTES:
    ...
EOF
   exit(0);
}

exit main();

# The main function
sub main {

    my ($me) = basename($0);

    # Process command-line arguments
    getopts ('c:h');

    print_syntax($me) if ($opt_h);

    $config = $opt_c if ($opt_c);
    parse_config($config);

    my %havedMl;   # stations that already have dMl while processing the list
    my $mergeOutFile = "merge_out";

    my $listFile = $ARGV[0];
    open IN, "<$listFile" or die "Can't open $listFile: $!\n";
  SN:
    while (<IN>) {
	chomp;
	my ($sn, $startDate) = split;

	if (-d $sn ) {
	    print STDERR "$sn already has a directory; skipping\n";
	    print "$sn already has a directory; skipping\n";
	    next SN;
	}

	my $cmd = "$bin/ml_events.pl -c $config -f $startDate -t $endDate -g $gridSpace -b $beforeTime -a $afterTime -s $sn > event.list";
	print STDERR "Running cmd: $cmd\n";
	open CMD, "$cmd |" or die "Can't run $cmd: $!\n";
	while (<CMD>) {
	    print STDERR $_;
	}
	close CMD or die "Error running $cmd: $!\n";
	$cmd = "/bin/wc -l event.list";
	open CMD, "$cmd |" or die "Can't run $cmd: $!\n";
	my $neq = 0;
	while (<CMD>) {
	    chomp;
	    if (/(\d+) /) {
		$neq = $1;
	    }
	}
	close CMD or die "Error running $cmd: $!\n";
	print STDERR "Got $neq events\n";
	if ($neq < $minEvents) {
	    print STDERR "Only $neq for $sn; skipping\n";
	    print "$sn none\n";
	    next SN;
	}

	print STDERR "Getting and merging amps for $sn\n";
	$cmd = "$bin/amp_merge.pl -c $config -B sncl_file -e event.list -s $sn -O $mergeOutFile 2>&1";
	open OUT, "> merge_was_stn_eq_data.screen" 
	    or die "Can't open merge_was_stn_eq_data.screen: $!\n";
	open CMD, "$cmd |" or die "Can't run $cmd: $!\n";
	while (<CMD>) {
	    print OUT $_;
	    if (/N_AF_OBS.*\s(\d+)\s.*_(\d+)_af_data/) {
		print STDERR "$2: $1 amps\n";
	    }
	}
	close CMD or die "Error running $cmd: $!";
	close OUT;

	print STDERR "Running dml_stats.pl\n";
	$cmd = "$bin/dml_stats.pl $mergeOutFile CISN_Candidate_SNCL_List > new_CISN_SNCL_dML.solution |& tee -";
	open CMD, "$cmd |" or die "Can't run $cmd: $!\n";
	while (<CMD>) {
	    print STDERR $_;
	}
	close CMD or die "Error running $cmd: $!\n";

	print STDERR "Running dml_thresh.pl\n";
	$cmd = "$bin/dml_thresh.pl -E $nEQ1 -e $nEQ2 -s $maxSE new_CISN_SNCL_dML.solution";
	open OUT, ">new_CISN_SNCL_dML.solution.thresh"
	    or die "Can't open new_CISN_SNCL_dML.solution.thresh: $!\n";
	open CMD, "$cmd |" or die "Can't run $cmd: $!\n";
	my $pass = 0;
	while (<CMD>) {
	    print OUT $_;
	    print $_;    # so it appears in doit.log
	    $pass = 1;
	}
	close OUT;
	close CMD or die "Error running $cmd: $!\n";
	if (-z "new_CISN_SNCL_dML.solution.thresh") {
	    unlink "new_CISN_SNCL_dML.solution.thresh"
		or die "Can't unlink empty new_CISN_SNCL_dML.solution.thresh: $!\n";
	    print "$sn none\n";
	}

	print STDERR "Making dir for $sn\n";
	my @mvFiles = ("event.list", "new_CISN_SNCL_dML.solution",
		       "new_CISN_SNCL_dML.solution.thresh",
		       "merge_was_stn_eq_data.screen", $mergeOutFile);
	mkdir $sn or die "Can't create $sn dir: $!\n";
	foreach my $file (@mvFiles) {
	    if (-e $file) {
		move $file, $sn or die "Can't move $file to $sn: $!\n";
	    }
	}

	# Clean up
	print STDERR "Cleaning up\n";
	if (-e $mergeOutFile) {
	    unlink $mergeOutFile or die "Can't delete $mergeOutFile: $!\n";
	}
    }
    close IN;

    return 0;
}

sub parse_config {
  my ($config) = @_;

  open CF, "< $config" or die "Error opening $config: $!\n";
  while (<CF>) {
      if (/^\s*END_DATE\s*=\s*(\S*)/) {
	  $endDate = $1;
      } elsif (/^\s*GRID_SPACE\s*=\s*(\S*)/) {
	  $gridSpace = $1;
      } elsif (/^\s*BEFORE_TIME\s*=\s*(\S*)/) {
	  $beforeTime = $1;
      } elsif (/^\s*AFTER_TIME\s*=\s*(\S*)/) {
	  $afterTime = $1;
      } elsif (/^\s*MIN_EVENTS\s*=\s*(\S*)/) {
	  $minEvents = $1;
      } elsif (/^\s*NEQ1\s*=\s*(\S*)/) {
	  $nEQ1 = $1;
      } elsif (/^\s*NEQ2\s*=\s*(\S*)/) {
	  $nEQ2 = $1;
      } elsif (/^\s*MAX_SE\s*=\s*(\S*)/) {
	  $maxSE = $1;
      } elsif (/^\s*BIN_DIR\s*=\s*(\S*)/) {
	  $bin = $1;
      }
  }
  close CF;

  die "BIN_DIR missing\n" unless (defined $bin);
  die "END_DATE missing\n" unless (defined $endDate);
  die "GRID_SPACE missing\n" unless (defined $gridSpace);
  die "BEFORE_TIME missing\n" unless (defined $beforeTime);
  die "AFTER_TIME missing\n" unless (defined $afterTime);
  die "MIN_EVENTS missing\n" unless (defined $minEvents);
  die "NEQ1 missing\n" unless (defined $nEQ1);
  die "NEQ2 missing\n" unless (defined $nEQ2);
  die "MAX_SE missing\n" unless (defined $maxSE);


  return;
}
