#!/usr/bin/perl

use strict;
use Getopt::Std;
use File::Basename;
use DBI;

# Changes:
# 2018/12/24: First version, PNL.
# 2021/08/02: Added logic to try events of lower magnitude within the
# lat-lon grid until one is found with no nearby events.
# 2021/08/06: Added option to include only events with given sta.net
# 2022/04/03: Added "o.rflag = 'F'" to query; makes end date less important
# 2022/12/02: Added upper magnitude limit, at request of Lind, Peggy, Gabbriele

# Global variables
our ( $VERSION, $config, $cmdname, $kpd, $maxMag, $mlAlgo);
our ( $opt_c, $opt_f, $opt_g, $opt_h, $opt_t, $opt_a, $opt_b, $opt_s );
our ( $dbname, $dbuser, $dbpass, $dbh );

$VERSION = "1.04 (2022/12/02)";
$config = "dMlAppend.conf";
$kpd = 111; # kilometers per degree
$| = 1;  # flush STDOUT

########################################################################
#	print_syntax	- print syntax and exit.
########################################################################
sub print_syntax {
    my ($cmdname) = @_;
    print
      "$cmdname version $VERSION
	$cmdname - Get list of Ml events for time period on grid
Syntax:
	$cmdname [-c config] [-h] [-f start -t end] -g grid-interval [-b sec] [-a sec] [-s sta.net]

where:
        -c config     - specify an alternate configuration file
                        The default config file is $config
	-h	Help - prints this help message.
	-f time -t end
		Specify start and end times to select events; required unless -s specified
	-g interval
		Specify  grid size, in km
	-s sta.net
		Specify a station and network which must provide WA amps to each event.

Notes:
	-f and -t options may be used with -s; without -s, they are required.
";
    exit(0);
}

    exit main();

sub main {

    $cmdname = basename($0);
    &getopts ('a:b:c:f:g:hs:t:');
    print_syntax($cmdname) if ($opt_h);
    print_syntax($cmdname) unless ($opt_g && ($opt_s || ($opt_f && $opt_t)));
    if ($opt_f && ! ($opt_f =~ m|^\d{4}/\d\d/\d\d$| &&
			 $opt_t =~ m|^\d{4}/\d\d/\d\d$|)) {
	print STDERR "malformed date strings\n";
	print_syntax($cmdname);
    }

    $config = $opt_c if ($opt_c);
    parse_config($config);

    my $checkNear = 0;
    my $checkBefore = 0;
    my $checkAfter = 0;
    if ($opt_b > 0 && $opt_a > 0) {
	$checkNear = 1;
	$checkBefore = $opt_b;
	$checkAfter = $opt_a;
    }
    my ($net, $sta);
    if ($opt_s) {
	($sta, $net) = split /\./, $opt_s;
	print_syntax($cmdname) unless (defined $sta && defined $net);
    }
    eval {
	$dbh = DBI->connect("dbi:Oracle:" . $dbname, $dbuser, $dbpass,
			    {RaiseError => 1,
			     PrintError => 0,
			     AutoCommit => 0});

	my $sql = q{
           SELECT UNIQUE e.evid, o.datetime, truetime.getstring(o.datetime) timestr,
           o.lat, o.lon, o.depth, n.magnitude,
           FLOOR((o.lat - 34.0) / :latsp) lati,
           FLOOR((o.lon + 126.0) / :lonsp) loni
           FROM event e, eventprefmag ep, origin o, netmag n};
	if ($opt_s) {
	    $sql .= q{, assocamm s, amp a};
	}

	$sql .= q{ WHERE e.evid = ep.evid AND e.prefor = o.orid
                   AND o.gtype != 't' AND o.rflag = 'F'};
	if ($opt_f) {
	    $sql .= " AND o.datetime BETWEEN truetime.putstring(:ostart)
            AND truetime.putstring(:oend)";
	}
	$sql .= q{ AND e.selectflag = 1 AND ep.magtype = 'l'
           AND ep.magid = n.magid AND n.magalgo = :mlalgo AND n.magnitude <= :maxmag};
	if ($opt_s) {
	    $sql .= q{ AND n.magid = s.magid AND s.ampid = a.ampid
                       AND a.net = :net AND a.sta = :sta};
	}
	$sql .= q{ ORDER BY lati, loni, n.magnitude DESC};
	my $evtSth = $dbh->prepare($sql);
	my $nearSth = $dbh->prepare(q{
           SELECT e.evid, truetime.getstring(o.datetime) datetime,
           o.lat, o.lon, n.magnitude
           FROM event e, origin o, netmag n
           WHERE e.prefor = o.orid AND e.prefmag = n.magid
           AND o.datetime BETWEEN :before AND :after
           AND e.selectflag = 1 AND e.evid != :nrevid});

	my $latsp = $opt_g / $kpd;
	my $lonsp = $opt_g / ($kpd * sin(36 * 3.14159/ 180));
	my $nevt = 1;

	if ($opt_f) {
	    $evtSth->bind_param(":ostart", $opt_f);
	    $evtSth->bind_param(":oend", $opt_t);
	}
	$evtSth->bind_param(":latsp", $latsp);
	$evtSth->bind_param(":lonsp", $lonsp);
	$evtSth->bind_param(":maxmag", $maxMag);
	$evtSth->bind_param(":mlalgo", $mlAlgo);
	if ($opt_s) {
	    $evtSth->bind_param(":net", $net);
	    $evtSth->bind_param(":sta", $sta);
	}
	$evtSth->execute();
	my $aryref = $evtSth->fetchall_arrayref({});
	die " no events found\n" unless (0 < scalar @$aryref);
	my $lastLatI = -100;
	my $lastLonI = -100;
	my $needEvent = 1;
	foreach my $ref (@$aryref) {
	    my $evid = $ref->{EVID};
	    my $datetime = $ref->{DATETIME};
	    my $timestr = $ref->{TIMESTR};
	    my $lat = $ref->{LAT};
	    my $lon = $ref->{LON};
	    my $depth = $ref->{DEPTH};
	    my $mag = $ref->{MAGNITUDE};
	    my $latI = $ref->{LATI};
	    my $lonI = $ref->{LONI};
#	    printf STDERR "DBG %8d %s %6.3f %8.3f %4.1f %3d %3d:\n",
#		$evid, $timestr, $lat, $lon, $mag, $latI,
#		    $lonI;
	    if ($latI != $lastLatI || $lonI != $lastLonI) {
		$needEvent = 1;
	    }
	    elsif ($needEvent == 0) {
		next;
	    }
	    if ($needEvent) {
		my $evtOK = 1;
		if ($checkNear) {
		    $nearSth->bind_param(":before", $datetime - $checkBefore);
		    $nearSth->bind_param(":after", $datetime + $checkAfter);
		    $nearSth->bind_param(":nrevid", $evid);
		    $nearSth->execute();
		    my $checkary = $nearSth->fetchall_arrayref({});
		    if (0 < scalar @$checkary) {
			printf STDERR "events near %03d %8d %s  %7.4f %9.4f %6.2f %4.1f ML NC %3d %3d:\n",
			    $nevt, $evid, $timestr, $lat, $lon,
				$depth, $mag, $latI, $lonI;
			foreach my $nref (@$checkary) {
			    printf STDERR "\t        %8d %s  %7.4f %9.4f %4.1f\n",
				$nref->{EVID}, $nref->{DATETIME},
				    $nref->{LAT}, $nref->{LON},
					$nref->{MAGNITUDE};
			}
			$evtOK = 0;
		    }
		}
		if ($evtOK) {
		    $lastLatI = $latI;
		    $lastLonI = $lonI;
		    printf "%8d %s  %7.4f %9.4f %6.2f %4.1f ML NC %3d %3d:\n",
			$evid, $timestr, $lat, $lon, $depth,
			    $mag, $latI, $lonI;
		    $nevt++;
		    $needEvent = 0;
		}
	    }
	}
    };
    if ($@) {
	print STDERR "error: $@\n";
    }
    $dbh->disconnect() if ($dbh);
    return 0;
}

sub parse_config {
  my ($config) = @_;

  open CF, "< $config" or die "Error opening $config: $!\n";
  while (<CF>) {
      if (/^\s*DB_USER\s*=\s*(\S+)/) {
	  $dbuser = $1;
      } elsif (/^\s*DB_PASSWORD\s*=\s*(\S*)/) {
	  $dbpass = $1;
      } elsif (/^\s*DB_NAME\s*=\s*(\S*)/) {
	  $dbname = $1;
      } elsif (/^\s*MAX_MAG\s*=\s*(\S*)/) {
	  $maxMag = $1;
      } elsif (/^\s*ML_ALGO\s*=\s*(\S*)/) {
	  $mlAlgo = $1;
      }
  }
  close CF;

  die "DB_USER missing\n" unless (defined $dbuser);
  die "DB_PASSWORD missing\n" unless (defined $dbpass);
  die "DB_NAME missing\n" unless (defined $dbname);
  die "MAX_MAG missing\n" unless (defined $maxMag);
  die "ML_ALGO missing\n" unless (defined $mlAlgo);

  return;
}
