#!/usr/bin/perl

# Query DB for stations without dMl values, based on file containing
# existing dMls

use strict;
use warnings;
use DBI;
use Getopt::Std;
use File::Basename;

our ($dbh, $dbname, $dbuser, $dbpass, $config, $VERSION, $mlAlgo);
our ($opt_c, $opt_h);

$config = "dMlAppend.conf";
$VERSION = "1.01 (2022/012/02)";

##############################################################################
#       print_syntax    - print syntax and exit.
##############################################################################
sub print_syntax {
    my ($cmdname) = @_;
    print STDERR << "EOF";
    $cmdname version $VERSION
        $cmdname - Query DB for stations without dMl values
Syntax:
        $cmdname  [-c config] [-h] dMl_sncl_list
where:
        -c config     - specify an alternate configuration file
                        The default config file is $config
	-h            - prints this help message.
	dml_sncl_list - file listing SNCLs for which dMl values are already known.

NOTES:
Writes to STDOUT.

    ...
EOF
   exit(0);
}

exit main();

# The main function
sub main {

    my ($me) = basename($0);

    # Process command-line arguments
    getopts ('c:h');

    print_syntax($me) if ($opt_h);
    print_syntax($me) unless (1 == scalar @ARGV);

    $config = $opt_c if ($opt_c);
    parse_config($config);

    my $dmlFile = $ARGV[0];
    my %dMl;

    # Make a hash of "sta.net" that have a dMl value, ignore component, location
    open IN, "<$dmlFile" or die "Can't open $dmlFile: $!\n";
    while (<IN>) {
	chomp;
	next if (/^\s*\#/); # skip comment lines
	my ($sncl, $dMl) = split;
	my ($s,$n,$c,$l) = split /\./, $sncl;
	my $sn = "$s.$n";
	$dMl{$sn} = 1;
    }
    close IN;

    eval {
	$dbh = DBI->connect("dbi:Oracle:" . $dbname, $dbuser, $dbpass,
			    {RaiseError => 1,
			     PrintError => 0,
			     AutoCommit => 0});
	my $sth = $dbh->prepare(q{
          SELECT a.net, a.sta, COUNT(e.evid) EVTS,
          to_char(MIN(cd.ondate), 'YYYY/MM/DD') ONDATE
          FROM event e, eventprefmag ep, assocamm s, amp a, channel_data cd,
          origin o, netmag n
          WHERE e.selectflag = 1 AND e.evid = ep.evid AND n.magalgo = :mlalgo
          AND e.prefor = o.orid AND o.rflag = 'F' AND ep.magtype = 'l'
          AND ep.magid = s.magid AND ep.magid = n.magid
          AND s.ampid = a.ampid AND a.net = cd.net AND a.sta = cd.sta
          AND a.seedchan = cd.seedchan AND a.location = cd.location
          AND to_date(truetime.getstring(a.datetime), 'YYYY/MM/DD HH24:MI:SS')
          BETWEEN cd.ondate AND cd.offdate
          GROUP BY a.net, a.sta ORDER BY EVTS DESC, a.net, a.sta});

	$sth->bind_param(":mlalgo", $mlAlgo);
	$sth->execute();
	my $aryref = $sth->fetchall_arrayref({});
	unless ($aryref && @$aryref) {
	    print STDERR "No SNCLs found\n";
	    $dbh->disconnect();
	    exit 0;
	}

	foreach my $ref (@$aryref) {
	    my $sta = $ref->{STA};
	    my $net = $ref->{NET};
	    my $nEvts = $ref->{EVTS};
	    my $start = $ref->{ONDATE};

	    my $sn = "$sta.$net";
	    next if (exists $dMl{$sn});
	    printf "%8s %s %5d\n", $sn, $start, $nEvts;
	}
    };
    if ($@) {
	print "error: $@\n";
    }
    $dbh->disconnect() if ($dbh);

    return 0;
}

sub parse_config {
    my ($config) = @_;

    open CF, "< $config" or die "Error opening $config: $!\n";
    while (<CF>) {
	if (/^\s*DB_USER\s*=\s*(\S+)/) {
	    $dbuser = $1;
	} elsif (/^\s*DB_PASSWORD\s*=\s*(\S*)/) {
	    $dbpass = $1;
	} elsif (/^\s*DB_NAME\s*=\s*(\S*)/) {
	    $dbname = $1;
	} elsif (/^\s*ML_ALGO\s*=\s*(\S*)/) {
	  $mlAlgo = $1;
	}
    }
    close CF;

    die "DB_USER missing\n" unless (defined $dbuser);
    die "DB_PASSWORD missing\n" unless (defined $dbpass);
    die "DB_NAME missing\n" unless (defined $dbname);
    die "ML_ALGO missing\n" unless (defined $mlAlgo);

    return;
}
