#!/usr/bin/perl

# Script to make map of events used for a given station's dMl

use strict;
use warnings;
use Getopt::Std;
use DBI;
use File::Basename;
use IPC::Run3;
use LogA0::CISN;

# Global variables
our($VERSION, $config );
our($opt_c,$opt_e,$opt_h,$opt_s,$opt_M, $opt_m);
our($dbname,$dbuser,$dbpass,$dbh,$plotUnit,$bin);

$config = "dMlAppend.conf";
$plotUnit = 'c';
$VERSION = "0.0.3 (2022/12/02)";

##############################################################################
#       print_syntax    - print syntax and exit.
##############################################################################
sub print_syntax {
    my ($cmdname) = @_;
    print STDERR << "EOF";
    $cmdname version $VERSION
        $cmdname - Generate map of events used for a given station
Syntax:
        $cmdname  [-h] [-c config] -e event-file [-m | -M merge_file] -s sta.net
where:
        -c config     - specify an alternate configuration file
                        The default config file is $config
        -e event-list - (required) file listing events
	-M merge-file - output file from amp_merge.pl
	-m	      - run amp_merge.pl, saving its output for use here
        -h            - prints this help message.
	-s SN         - station.net for which dMl values are desired.

NOTES:
Output postscript maps will be named station.net_E.ps and station.net_N.ps
    ...
EOF
    exit(0);
}


    exit main();

# The main function
sub main {

    my ($me) = basename($0);

    # Process command-line arguments
    getopts ('c:e:hM:ms:');

    print_syntax($me) if ($opt_h);
    unless ($opt_e) {
	print STDERR "error: no event list given\n";
	print_syntax($me);
    }
    my $evtFile = $opt_e;
    unless ($opt_M || $opt_m) {
	print STDERR "error: no merge option given\n";
	print_syntax($me);
    }
    my $mergeFile;
    if ($opt_M) {
	$mergeFile = $opt_M;
    }
    unless ($opt_s) {
	print STDERR "error: sta.net not given\n";
	print_syntax($me);
    }
    my $sn = $opt_s;
    if ($opt_m) {
	$mergeFile = runMerge( $sn );
    }

    $config = $opt_c if ($opt_c);
    parse_config($config);

    my (%events, %range);
    parseEventFile($evtFile, \%events);
    parseMergeFile($sn, $mergeFile, \%events);
    @{ $range{COUNT} } = getCountLimits(\%events);

    my $logA0 = LogA0::CISN->new();
    @{ $range{MAG} } = getMagLimits(\%events);
    our ($minRes, $maxRes) = (10, -10);
    our ($minLat, $maxLat) = (90, -90);
    our ($minLon, $maxLon) = (180, -180);

    eval {
	$dbh = DBI->connect("dbi:Oracle:" . $dbname, $dbuser, $dbpass,
			    {RaiseError => 1,
			     PrintError => 0,
			     AutoCommit => 0});

	my $sth = $dbh->prepare(q{
       SELECT a.amplitude, a.amptype, a.units, wheres.separation_km(o.lat, o.lon,
       -o.depth, c.lat, c.lon, c.elev / 1000) DIST, c.lat, c.lon
       FROM event e, eventprefmag ep, origin o, assocamm s, amp a, channel_data c
       WHERE e.evid = :evid AND e.prefor = o.orid AND e.evid = ep.evid 
       AND ep.magtype = 'l' AND ep.magid = s.magid
       AND s.ampid = a.ampid AND a.net = :net AND a.sta = :sta AND a.seedchan = :chan
       AND a.location = :loc AND a.net = c.net AND a.sta = c.sta
       AND a.seedchan = c.seedchan AND a.location = c.location
       AND to_date(truetime.getstring(a.datetime), 'YYYY/MM/DD HH24:MI:SS')
       BETWEEN c.ondate AND c.offdate});

	open MAGE, ">magE.dat" or die "Can't open magE.dat: $!\n";
	open MAGN, ">magN.dat" or die "Can't open magN.dat: $!\n";
	open RESE, ">resE.dat" or die "Can't open resE.dat: $!\n";
	open RESN, ">resN.dat" or die "Can't open resN.dat: $!\n";

	my ($staLat, $staLon);
	foreach my $evid (sort keys %events) {
	    foreach my $sncl (keys %{ $events{$evid} }) {
		next unless ($sncl =~ /\./);
		my ($sta, $net, $chan, $loc) = split /\./, $sncl;
		$loc =~ s/-/ /g;
		my $dir = substr $chan, 2, 1;
		if ($dir eq "E") {
		    printf MAGE "%10.5f %8.5f %5.2f %5.2f%s\n", $events{$evid}->{LON},
			$events{$evid}->{LAT}, $events{$evid}->{MAG},
			    countToSize($events{$evid}->{$sncl}->{COUNT}, $range{COUNT}), $plotUnit;
		} elsif ($dir eq "N") {
		    printf MAGN "%10.5f %8.5f %5.2f %5.2f%s\n", $events{$evid}->{LON},
			$events{$evid}->{LAT}, $events{$evid}->{MAG},
			    countToSize($events{$evid}->{$sncl}->{COUNT}, $range{COUNT}), $plotUnit;
		}

		$sth->bind_param(":evid", $evid);
		$sth->bind_param(":sta", $sta);
		$sth->bind_param(":net", $net);
		$sth->bind_param(":chan", $chan);
		$sth->bind_param(":loc", $loc);
		$sth->execute();
		my $aryref = $sth->fetchall_arrayref({});
		unless (0 < scalar @$aryref) {
		    print STDERR "Nothing in DB for $sncl and $evid\n";
		    next;
		}
		my $amp = $aryref->[0]->{AMPLITUDE};
		my $units = $aryref->[0]->{UNITS};
		my $dist = $aryref->[0]->{DIST};
		$staLat = $aryref->[0]->{LAT} unless (defined $staLat);
		$staLon = $aryref->[0]->{LON} unless (defined $staLon);

		if ($units eq "cm") {
		    $amp *= 10;
		}
		my $resid = sprintf "%5.2f",
		    getMag($amp, $logA0, $dist) - $events{$evid}->{MAG};
		$minRes = $resid if ($minRes > $resid);
		$maxRes = $resid if ($maxRes < $resid);

		if ($dir eq "E") {
		    printf RESE "%10.5f %8.5f %5.2f %5.2f%s\n", $events{$evid}->{LON},
			$events{$evid}->{LAT}, $resid,
			    countToSize($events{$evid}->{$sncl}->{COUNT}, $range{COUNT}),
				$plotUnit;
		} elsif ($dir eq "N") {
		    printf RESN "%10.5f %8.5f %5.2f %5.2f%s\n", $events{$evid}->{LON},
			$events{$evid}->{LAT}, $resid,
			    countToSize($events{$evid}->{$sncl}->{COUNT}, $range{COUNT}),
				$plotUnit;
		}
		$minLat = $events{$evid}->{LAT} if ($minLat > $events{$evid}->{LAT});
		$maxLat = $events{$evid}->{LAT} if ($maxLat < $events{$evid}->{LAT});
		$minLon = $events{$evid}->{LON} if ($minLon > $events{$evid}->{LON});
		$maxLon = $events{$evid}->{LON} if ($maxLon < $events{$evid}->{LON});
	    }
	}
	close MAGE;
	close MAGN;
	close RESE;
	close RESN;

	$minLat = $staLat if ($minLat > $staLat);
	$maxLat = $staLat if ($maxLat < $staLat);
	$minLon = $staLon if ($minLon > $staLon);
	$maxLon = $staLon if ($maxLon < $staLon);
	@{ $range{RES} } = ($minRes, $maxRes);
	@{ $range{LAT} } = ($minLat, $maxLat);
	@{ $range{LON} } = ($minLon, $maxLon);
	my %sta = (LAT => $staLat,
		   LON => $staLon,
		  );
	print "  magLimits: ", join(" ", @{ $range{MAG} }), "\n";
	print "  resLimits: ", join(" ", @{ $range{RES} }), "\n";
	print "countLimits: ", join( " ", @{ $range{COUNT} }), "\n";
	print "  latLimits: ", join(" ", @{ $range{LAT} }), "\n";
	print "  lonLimits: ", join(" ", @{ $range{LON} }), "\n";

	makePlot($sn, "E", \%range, \%sta);
	makePlot($sn, "N", \%range, \%sta);

    };
    if ($@) {
	print STDERR "Error: $@\n";
    }
    if ($dbh) {
	$dbh->disconnect();
    }


    return 0;
}

sub parseEventFile {
    my ($evtFile, $ref) = @_;

    open FH, "<$evtFile" or die "Can'topen $evtFile: $!\n";
    while (<FH>) {
	chomp;
	my @ary = split;
	my $evid = $ary[0];
	my $lat = $ary[2];
	my $lon = $ary[3];
	my $mag = $ary[5];
	$ref->{$evid} = {LAT => $lat,
			 LON => $lon,
			 MAG => $mag,
		     };
    }
    close FH;

    return;
}

sub parseMergeFile {
    my ($sn, $mergeFile, $ref) = @_;

    open FH, "<$mergeFile" or die "Can't open $mergeFile: $!\n";
    while (<FH>) {
	chomp;
	my @ary =split;
	my $evid = $ary[0];
	my $sncl = $ary[1];
	$ref->{$evid}->{$sncl}->{COUNT} = 0 unless (exists $ref->{$evid}->{$sncl});
	$ref->{$evid}->{$sncl}->{COUNT}++;
    }
    close FH;

    return;
}

sub getCountLimits {
    my $ref = shift;

    my ($min, $max) = (1.0e+6,0);
    foreach my $evid (keys %$ref) {
	foreach my $sncl (keys %{ $ref->{$evid} }) {
	    next unless ($sncl =~ /\./);
	    $max = $ref->{$evid}->{$sncl}->{COUNT}
		if ($ref->{$evid}->{$sncl}->{COUNT} > $max);
	    $min = $ref->{$evid}->{$sncl}->{COUNT}
		if ($ref->{$evid}->{$sncl}->{COUNT} > 0 &&
			$ref->{$evid}->{$sncl}->{COUNT} < $min);
	}
    }

    return ($min, $max);
}

# Compute Ml magnitude for the best CL of each SN-Dir.
sub getMag {
    my ($amp, $logA0, $dist) = @_;

    my $ml = log10($amp) + $logA0->logA0($dist);

    return $ml;
}


# Convert obervation counts to size for psxy.
sub countToSize {
    my $count = shift;
    my $ref = shift;
    my $minSize = 0.1;
    my $maxSize = 0.5;

    # Make the circle area proportional to counts. This might make the
    # smallest circles pretty tiny...
    my $size = $maxSize * sqrt($count) / sqrt($ref->[1]);

    return $size;
}

sub log10 {
  my ($arg) = @_;

  return log($arg) / log(10);
}

sub getMagLimits {
    my $ref = shift;

    my ($min, $max) = (10, -10);
    foreach my $evid (keys %$ref ) {
	if (exists $ref->{$evid}->{MAG}) {
	    $min = $ref->{$evid}->{MAG} if ($min > $ref->{$evid}->{MAG});
	    $max = $ref->{$evid}->{MAG} if ($max < $ref->{$evid}->{MAG});
	}
    }

    return ($min, $max);
}

# Run amp_merge.pl to generate merge output file, which this program needs.
# Does not save amp_merge.pl's STDERR, assumed to be already saved by doit.pl
# Assumes event.list file is in current directory
# Throws an exception if run3 or amp_merge.pl failed.
# Otherwise returns the name of amp_merge.pl's output.
sub runMerge {
    my $sn = shift;

    my $conf = "$bin/amp_merge.conf";
    my $basis = "$bin/../sncl_file";
    my $eventFile = "event.list";
    my $mergeOut = "merge_out";
    my @cmd = ("$bin/amp_merge.pl", "-c", "$conf", "-B", "$basis", "-e",
	       "$eventFile", "-s", "$sn", "-O", "$mergeOut");
    running( \@cmd );
    run3 \@cmd, \undef, \undef, \undef; # Redirect all output to /dev/null
    my $res = $?;
    checkRes($res);

    return $mergeOut;
}

# Make plots for each direction using GMT commands.
# Arguments are:
#   STA.NET
#   direction, e.g. E or N
#   hashref of limits
#   hashref of station coordinates LAT, LON
sub makePlot {
    my $sn = shift;
    my $dir = shift;
    my $lim = shift;
    my $sta = shift;

    my $mFile = "mag$dir.dat";
    my $rFile = "res$dir.dat";
    my $outFile = "${sn}_$dir.ps";

    # GMT parameters:
#    my $R = sprint("-R%.2f/%.2f/%.2f/%.2f", $lim->{LON}->[0], $lim->{LON}->[1],
# 		   $lim->{LAT}->[0], $lim->{LAT}->[1]);
    # Above -R param will mess with scale and text location, as well as map aspect.
    # For now, use a fixed projection:
    my $R = "-R-128/-114/31/44";
    my $I = "-I5m";
    my $J = "-JM7i";
    my $mPal = "mag.cpt";
    my $PM = "-C$mPal"; # mag palette
    my $rPal = "res.cpt";
    my $PR = "-C$rPal"; # residual palette
    my $Y = "-Y1i";
    my $B = "-Ba2f2WSne"; # psbasemap parameter
    my $staSymbol = "-Sa0.5c";
    my $staColor = "-Ggreen";
    my $magSymbol = "-Skucirc";
    my $resSymbol = "-Sklcirc";
    my $DMag = "-D0.2i/4.2i/2i/0.2i"; # psscale parameter for magnitude
    my $DRes = "-D0.2i/1.3i/2i/0.2i"; # psscale parameter for residual
    my $TF = 6; # pstext font
    my $W = "-Wthick";  # circumference pen width
    my $Bmag = "-B0.5";
    my $Bres = "-B0.25";
    my $scaleFontSize = 10;
    my $headFontSize = 16;
    my $srcMagPalette = "-Cpolar";    # red2blue standard GMT palette
    my $srcResPalette = "-Crainbow";  # standard GMT palette

    my (@cmd);

    # Make the palettes if they don't already exist
    unless (-e $mPal) {
 	my $T = sprintf "-T%.2f/%.2f/%.3f", $lim->{MAG}->[0], $lim->{MAG}->[1],
 	    ($lim->{MAG}->[1] - $lim->{MAG}->[0]) * 0.5;
 	@cmd = ("makecpt", "$srcMagPalette", $T, "-Z");
	running(\@cmd);
 	run3 \@cmd, \undef, $mPal, undef;  # write to PM file, STDERR displayed
	my $res = $?;
	checkRes($res);
     }
     unless (-e $rPal) {
 	my $T = sprintf "-T%.2f/%.2f/%.3f", $lim->{RES}->[0], $lim->{RES}->[1],
 	    ($lim->{RES}->[1] - $lim->{RES}->[0]) * 0.5;
 	@cmd = ("makecpt", "$srcResPalette", $T, "-Z");
	running(\@cmd);
	run3 \@cmd, \undef, $rPal, undef; 
	my $res = $?;
	checkRes($res);
    }

    open(my $psFH, ">", $outFile) or die "Can't create $outFile: $!\n";
    # basemap
    @cmd = ("psbasemap", $J, $R, $Y, $B, "-P", "-K");
    running(\@cmd);
    run3 \@cmd, \undef, $psFH, undef;
    my $res = $?;
    checkRes($res);

    # coast and landmass outline
    @cmd = ("pscoast", $J, $R, "-Dh", "-W1", "-N1", "-N2", "-O", "-K");
    running(\@cmd);
    run3 \@cmd, \undef, $psFH, undef;
    $res = $?;
    checkRes($res);

    # Station symbol
    @cmd = ("psxy", $J, $R, $staSymbol, "-O", "-K", $staColor);
    running(\@cmd);
    my $staText = sprintf "%.5f %.5f\n", $sta->{LON}, $sta->{LAT};
    run3 \@cmd, \$staText, $psFH, undef;
    $res = $?;
    checkRes($res);

    # Magnitude in upper half circle
    @cmd = ("psxy", $mFile, $J, $R, "-P", $magSymbol, $PM, "-O", "-K");
    running(\@cmd);
    run3 \@cmd, \undef, $psFH, undef;
    $res = $?;
    checkRes($res);

    # Scale and text for magnitude
    @cmd = ("psscale", $DMag, $Bmag, "-O", "-K", "-P", $PM);
    running(\@cmd);
    run3 \@cmd, \undef, $psFH, undef;
    $res = $?;
    checkRes($res);
    @cmd = ("pstext", $J, $R, "-O", "-K");
    running(\@cmd);
    my $magScaleText = "-126.5 36.1.0 $scaleFontSize 0.0 0 $TF Magnitude (upper half)\n";
    run3 \@cmd, \$magScaleText, $psFH, undef;
    $res = $?;
    checkRes($res);

    # Residual in lower half circle
    @cmd = ("psxy", $rFile, $J, $R, "-P", $resSymbol, $PR, "-O", "-K");
    running(\@cmd);
    run3 \@cmd, \undef, $psFH, undef;
    $res = $?;
    checkRes($res);

    # Scale and text for residual
    @cmd = ("psscale", $DRes, $Bres, "-O", "-K", "-P", $PR);
    running(\@cmd);
    run3 \@cmd, \undef, $psFH, undef;
    $res = $?;
    checkRes($res);
    @cmd = ("pstext", $J, $R, "-O", "-K");
    running(\@cmd);
    my $resScaleText = "-126.5 31.2 $scaleFontSize 0.0 0 $TF Residual (lower half)\n";
    run3 \@cmd, \$resScaleText, $psFH, undef;
    $res = $?;
    checkRes($res);

    # Header text, last GMT command for this file.
    @cmd = ("pstext", $J, $R, "-N", "-O");
    my $headText = sprintf("-121.0 44.5 %d 0.0 0 %d Event Contributions for %s %s",
			   $headFontSize, $TF, $sn, $dir);
    running(\@cmd);
    run3 \@cmd, \$headText, $psFH, undef;
    $res = $?;
    checkRes($res);

    close $psFH or die "Can't close $outFile: $!\n";
}

# Prints a statement about the command (arrayref argument) that is about to
# be run.
sub running {
    my $cmd = shift;

    print STDERR "running ", join( " ", @$cmd), "\n";
}

# Checks the value of $? which has been passed as the only argument.
# On errors, die's with appropriate message
sub checkRes {

    my $res = shift;
    return $res unless ($res);

    if ($? == -1) {
        die "error: failed to execute: $!\n";
    }
    elsif ($? & 127) {
        die( sprintf "error: child died with signal %d, %s coredump\n",
            ($? & 127),  ($? & 128) ? 'with' : 'without');
    }
    else {
	my $status = $? >> 8;
        die "Command returned $status\n";
    }
    # never get here

}

sub parse_config {
  my ($config) = @_;

  open CF, "< $config" or die "Error opening $config: $!\n";
  while (<CF>) {
      if (/^\s*DB_USER\s*=\s*(\S+)/) {
	  $dbuser = $1;
      } elsif (/^\s*DB_PASSWORD\s*=\s*(\S*)/) {
	  $dbpass = $1;
      } elsif (/^\s*DB_NAME\s*=\s*(\S*)/) {
	  $dbname = $1;
      } elsif (/^\s*BIN_DIR\s*=\s*(\S*)/) {
	  $bin = $1;
      }
  }
  close CF;

  die "DB_USER missing\n" unless (defined $dbuser);
  die "DB_PASSWORD missing\n" unless (defined $dbpass);
  die "DB_NAME missing\n" unless (defined $dbname);
  die "BIN_DIR missing\n" unless (defined $bin);

  return;
}
